@@ -740,7 +740,7 @@ void Node::ncSendError(uint64_t nwid,uint64_t requestPacketId,const Address &des
 				n->setAccessDenied(nullptr);
 				break;
 			case NetworkController::NC_ERROR_AUTHENTICATION_REQUIRED: {
-				fprintf(stderr, "\n\nGot auth required\n\n");
+				//fprintf(stderr, "\n\nGot auth required\n\n");
 
 				break;
 			} 
