@@ -13,11 +13,24 @@
 use thiserror::Error;
 
 #[derive(Error, Debug)]
-pub enum ZeroIDCError
-{
+pub enum ZeroIDCError {
     #[error(transparent)]
-    DiscoveryError(#[from] openidconnect::DiscoveryError<openidconnect::reqwest::Error<reqwest::Error>>),
+    DiscoveryError(
+        #[from] openidconnect::DiscoveryError<openidconnect::reqwest::Error<reqwest::Error>>,
+    ),
 
     #[error(transparent)]
     ParseError(#[from] url::ParseError),
 }
+
+#[derive(Error, Debug)]
+#[error("SSO Exchange Error: {message:}")]
+pub struct SSOExchangeError {
+    message: String,
+}
+
+impl SSOExchangeError {
+    pub fn new(message: String) -> Self {
+        SSOExchangeError { message }
+    }
+}
